"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.zodTypeToTsTypeDef = exports.serializeZodType = exports.unwrapPromiseOnce = exports.apiParamTypeToTsType = exports.apiParamToTsCode = exports.zodTypeToZodCode = exports.propToZodTypeDef = exports.getTextTypeLength = exports.propNodeToZodTypeDef = exports.getZodTypeFromApiParamType = exports.getZodObjectFromApiParams = exports.getZodObjectFromApi = void 0;
const zod_1 = require("zod");
const types_1 = require("../types/types");
/*
  ExtendedApi 에서 ZodObject 리턴
*/
function getZodObjectFromApi(api, references = {}) {
    var _a;
    if (((_a = api.typeParameters) === null || _a === void 0 ? void 0 : _a.length) > 0) {
        api.typeParameters.map((typeParam) => {
            if (typeParam.constraint) {
                let zodType = getZodTypeFromApiParamType(typeParam.constraint, references);
                references[typeParam.id] = zodType;
            }
        });
    }
    const ReqType = getZodObjectFromApiParams(api.parameters.filter((param) => !types_1.ApiParamType.isContext(param.type) &&
        !types_1.ApiParamType.isRefKnex(param.type)), references);
    return ReqType;
}
exports.getZodObjectFromApi = getZodObjectFromApi;
/*
  ZodObject를 통해 ApiParam 리턴
*/
function getZodObjectFromApiParams(apiParams, references = {}) {
    return zod_1.z.object(apiParams.reduce((r, param) => {
        let zodType = getZodTypeFromApiParamType(param.type, references);
        if (param.optional) {
            zodType = zodType.optional();
        }
        return Object.assign(Object.assign({}, r), { [param.name]: zodType });
    }, {}));
}
exports.getZodObjectFromApiParams = getZodObjectFromApiParams;
/*
  ApiParamType으로 ZodType 컨버팅
*/
function getZodTypeFromApiParamType(paramType, references) {
    var _a, _b;
    switch (paramType) {
        case "string":
            return zod_1.z.string();
        case "number":
            return zod_1.z.number();
        case "boolean":
            return zod_1.z.boolean();
        default:
            const advType = paramType;
            switch (advType.t) {
                case "string-literal":
                case "numeric-literal":
                    return zod_1.z.literal(advType.value);
                case "object":
                    const objType = paramType;
                    return getZodObjectFromApiParams(objType.props);
                case "array":
                    const arrType = paramType;
                    return zod_1.z.array(getZodTypeFromApiParamType(arrType.elementsType, references));
                case "ref":
                    const refType = paramType;
                    // 객체 키 관리 유틸리티
                    if (["Pick", "Omit"].includes(refType.id)) {
                        if (((_a = refType.args) === null || _a === void 0 ? void 0 : _a.length) !== 2) {
                            throw new Error(`잘못된 ${refType.id}`);
                        }
                        const [obj, literalOrUnion] = refType.args.map((arg) => getZodTypeFromApiParamType(arg, references));
                        let keys = [];
                        if (literalOrUnion instanceof zod_1.z.ZodUnion) {
                            keys = literalOrUnion._def.options.map((option) => option._def.value);
                        }
                        else {
                            keys = [literalOrUnion._def.value];
                        }
                        const keyRecord = keys.reduce((result, key) => {
                            return Object.assign(Object.assign({}, result), { [key]: true });
                        }, {});
                        if (refType.id === "Pick") {
                            if (obj.pick) {
                                return obj.pick(keyRecord);
                            }
                        }
                        else {
                            if (obj.omit) {
                                return obj.omit(keyRecord);
                            }
                        }
                    }
                    if (["Partial"].includes(refType.id)) {
                        if (((_b = refType.args) === null || _b === void 0 ? void 0 : _b.length) !== 1) {
                            throw new Error(`잘못된 ${refType.id}`);
                        }
                        const obj = getZodTypeFromApiParamType(refType.args[0], references);
                        return obj.partial();
                    }
                    const reference = references[refType.id];
                    if (reference === undefined) {
                        return zod_1.z.string();
                        // throw new Error(`ref 참조 불가 ${refType.id}`);
                    }
                    return reference;
                case "union":
                    const unionType = paramType;
                    return zod_1.z.union(unionType.types.map((type) => getZodTypeFromApiParamType(type, references)));
                case "intersection":
                    const intersectionType = paramType;
                    return intersectionType.types.reduce((result, type, index) => {
                        const resolvedType = getZodTypeFromApiParamType(type, references);
                        if (index === 0) {
                            return resolvedType;
                        }
                        else {
                            return zod_1.z.intersection(result, resolvedType);
                        }
                    }, zod_1.z.unknown());
                case "tuple-type":
                    const tupleType = paramType;
                    return zod_1.z.tuple(tupleType.elements.map((elem) => getZodTypeFromApiParamType(elem, references)));
            }
            return zod_1.z.unknown();
    }
}
exports.getZodTypeFromApiParamType = getZodTypeFromApiParamType;
function propNodeToZodTypeDef(propNode, injectImportKeys) {
    if (propNode.nodeType === "plain") {
        return propToZodTypeDef(propNode.prop, injectImportKeys);
    }
    else if (propNode.nodeType === "array") {
        return [
            propNode.prop ? `${propNode.prop.name}: ` : "",
            "z.array(z.object({",
            propNode.children
                .map((childPropNode) => propNodeToZodTypeDef(childPropNode, injectImportKeys))
                .join("\n"),
            "",
            "})),",
        ].join("\n");
    }
    else if (propNode.nodeType === "object") {
        return [
            propNode.prop ? `${propNode.prop.name}: ` : "",
            "z.object({",
            propNode.children
                .map((childPropNode) => propNodeToZodTypeDef(childPropNode, injectImportKeys))
                .join("\n"),
            "",
            `})${propNode.prop && propNode.prop.nullable ? ".nullable()" : ""},`,
        ].join("\n");
    }
    else {
        throw Error;
    }
}
exports.propNodeToZodTypeDef = propNodeToZodTypeDef;
function getTextTypeLength(textType) {
    switch (textType) {
        case "text":
            return 1024 * 64 - 1;
        case "mediumtext":
            return 1024 * 1024 * 16 - 1;
        case "longtext":
            return 1024 * 1024 * 1024 * 4 - 1;
    }
}
exports.getTextTypeLength = getTextTypeLength;
function propToZodTypeDef(prop, injectImportKeys) {
    let stmt;
    if ((0, types_1.isIntegerProp)(prop)) {
        stmt = `${prop.name}: z.number().int()`;
    }
    else if ((0, types_1.isBigIntegerProp)(prop)) {
        stmt = `${prop.name}: z.bigint()`;
    }
    else if ((0, types_1.isTextProp)(prop)) {
        stmt = `${prop.name}: z.string().max(${getTextTypeLength(prop.textType)})`;
    }
    else if ((0, types_1.isEnumProp)(prop)) {
        stmt = `${prop.name}: ${prop.id}`;
        injectImportKeys.push(prop.id);
    }
    else if ((0, types_1.isStringProp)(prop)) {
        stmt = `${prop.name}: z.string().max(${prop.length})`;
    }
    else if ((0, types_1.isDecimalProp)(prop)) {
        stmt = `${prop.name}: z.string()`;
    }
    else if ((0, types_1.isFloatProp)(prop) || (0, types_1.isDoubleProp)(prop)) {
        stmt = `${prop.name}: z.number()`;
    }
    else if ((0, types_1.isBooleanProp)(prop)) {
        stmt = `${prop.name}: z.boolean()`;
    }
    else if ((0, types_1.isDateProp)(prop)) {
        stmt = `${prop.name}: z.string().length(10)`;
    }
    else if ((0, types_1.isTimeProp)(prop)) {
        stmt = `${prop.name}: z.string().length(8)`;
    }
    else if ((0, types_1.isDateTimeProp)(prop)) {
        stmt = `${prop.name}: SQLDateTimeString`;
    }
    else if ((0, types_1.isTimestampProp)(prop)) {
        stmt = `${prop.name}: SQLDateTimeString`;
    }
    else if ((0, types_1.isJsonProp)(prop)) {
        stmt = `${prop.name}: ${prop.id}`;
        injectImportKeys.push(prop.id);
    }
    else if ((0, types_1.isUuidProp)(prop)) {
        stmt = `${prop.name}: z.string().uuid()`;
    }
    else if ((0, types_1.isVirtualProp)(prop)) {
        stmt = `${prop.name}: ${prop.id}`;
        injectImportKeys.push(prop.id);
    }
    else if ((0, types_1.isRelationProp)(prop)) {
        if ((0, types_1.isBelongsToOneRelationProp)(prop) ||
            ((0, types_1.isOneToOneRelationProp)(prop) && prop.hasJoinColumn)) {
            stmt = `${prop.name}_id: z.number().int()`;
        }
        else {
            // 그외 relation 케이스 제외
            return `// ${prop.name}: ${prop.relationType} ${prop.with}`;
        }
    }
    else {
        return "// unable to resolve";
    }
    if (prop.unsigned) {
        stmt += ".nonnegative()";
    }
    if (prop.nullable) {
        stmt += ".nullable()";
    }
    return stmt + ",";
}
exports.propToZodTypeDef = propToZodTypeDef;
function zodTypeToZodCode(zt) {
    switch (zt._def.typeName) {
        case "ZodString":
            return "z.string()";
        case "ZodNumber":
            return "z.number()";
        case "ZodBoolean":
            return "z.boolean()";
        case "ZodBigInt":
            return "z.bigint()";
        case "ZodDate":
            return "z.date()";
        case "ZodNull":
            return "z.null()";
        case "ZodUndefined":
            return "z.undefined()";
        case "ZodAny":
            return "z.any()";
        case "ZodUnknown":
            return "z.unknown()";
        case "ZodNever":
            return "z.never()";
        case "ZodNullable":
            return zodTypeToZodCode(zt._def.innerType) + ".nullable()";
        case "ZodDefault":
            return (zodTypeToZodCode(zt._def.innerType) +
                `.default(${zt._def.defaultValue()})`);
        case "ZodRecord":
            return `z.record(${zodTypeToZodCode(zt._def.keyType)}, ${zodTypeToZodCode(zt._def.valueType)})`;
        case "ZodLiteral":
            if (typeof zt._def.value === "string") {
                return `z.literal("${zt._def.value}")`;
            }
            else {
                return `z.literal(${zt._def.value})`;
            }
        case "ZodUnion":
            return `z.union([${zt._def.options
                .map((option) => zodTypeToZodCode(option))
                .join(",")}])`;
        case "ZodEnum":
            return `z.enum([${zt._def.values
                .map((val) => `"${val}"`)
                .join(", ")}])`;
        case "ZodArray":
            return `z.array(${zodTypeToZodCode(zt._def.type)})`;
        case "ZodObject":
            const shape = zt.shape;
            return [
                "z.object({",
                ...Object.keys(shape).map((key) => `${key}: ${zodTypeToZodCode(shape[key])},`),
                "})",
            ].join("\n");
        case "ZodOptional":
            return zodTypeToZodCode(zt._def.innerType) + ".optional()";
        default:
            throw new Error(`처리되지 않은 ZodType ${zt._def.typeName}`);
    }
}
exports.zodTypeToZodCode = zodTypeToZodCode;
function apiParamToTsCode(params, injectImportKeys) {
    return params
        .map((param) => {
        return `${param.name}${param.optional && !param.defaultDef ? "?" : ""}: ${apiParamTypeToTsType(param.type, injectImportKeys)}${param.defaultDef ? `= ${param.defaultDef}` : ""}`;
    })
        .join(", ");
}
exports.apiParamToTsCode = apiParamToTsCode;
function apiParamTypeToTsType(paramType, injectImportKeys) {
    if ([
        "string",
        "number",
        "boolean",
        "true",
        "false",
        "null",
        "undefined",
        "void",
        "any",
        "unknown",
    ].includes(paramType)) {
        return paramType;
    }
    else if (types_1.ApiParamType.isObject(paramType)) {
        return `{ ${apiParamToTsCode(paramType.props, injectImportKeys)} }`;
    }
    else if (types_1.ApiParamType.isStringLiteral(paramType)) {
        return `"${paramType.value}"`;
    }
    else if (types_1.ApiParamType.isNumericLiteral(paramType)) {
        return String(paramType.value);
    }
    else if (types_1.ApiParamType.isUnion(paramType)) {
        return paramType.types
            .map((type) => apiParamTypeToTsType(type, injectImportKeys))
            .join(" | ");
    }
    else if (types_1.ApiParamType.isIntersection(paramType)) {
        return paramType.types
            .map((type) => apiParamTypeToTsType(type, injectImportKeys))
            .join(" & ");
    }
    else if (types_1.ApiParamType.isArray(paramType)) {
        return (apiParamTypeToTsType(paramType.elementsType, injectImportKeys) + "[]");
    }
    else if (types_1.ApiParamType.isRef(paramType)) {
        if (["Pick", "Omit", "Promise", "Partial"].includes(paramType.id) === false) {
            // importKeys 인젝션
            injectImportKeys.push(paramType.id);
        }
        if (paramType.args === undefined || paramType.args.length === 0) {
            return paramType.id;
        }
        else {
            return `${paramType.id}<${paramType.args
                .map((arg) => apiParamTypeToTsType(arg, injectImportKeys))
                .join(",")}>`;
        }
    }
    else if (types_1.ApiParamType.isIndexedAccess(paramType)) {
        return `${apiParamTypeToTsType(paramType.object, injectImportKeys)}[${apiParamTypeToTsType(paramType.index, injectImportKeys)}]`;
    }
    else if (types_1.ApiParamType.isTupleType(paramType)) {
        return `[ ${paramType.elements.map((elem) => apiParamTypeToTsType(elem, injectImportKeys))} ]`;
    }
    else if (types_1.ApiParamType.isTypeParam(paramType)) {
        return `<${paramType.id}${paramType.constraint
            ? ` extends ${apiParamTypeToTsType(paramType.constraint, injectImportKeys)}`
            : ""}>`;
    }
    else {
        throw new Error(`resolve 불가 ApiParamType ${paramType}`);
    }
}
exports.apiParamTypeToTsType = apiParamTypeToTsType;
function unwrapPromiseOnce(paramType) {
    if (types_1.ApiParamType.isPromise(paramType)) {
        return paramType.args[0];
    }
    else {
        return paramType;
    }
}
exports.unwrapPromiseOnce = unwrapPromiseOnce;
function serializeZodType(zt) {
    switch (zt._def.typeName) {
        case "ZodObject":
            return {
                type: "object",
                shape: Object.keys(zt.shape).reduce((result, key) => {
                    return Object.assign(Object.assign({}, result), { [key]: serializeZodType(zt.shape[key]) });
                }, {}),
            };
        case "ZodArray":
            return {
                type: "array",
                element: serializeZodType(zt._def.type),
            };
        case "ZodEnum":
            return {
                type: "enum",
                values: zt._def.values,
            };
        case "ZodString":
            return {
                type: "string",
                checks: zt._def.checks,
            };
        case "ZodNumber":
            return {
                type: "number",
                checks: zt._def.checks,
            };
        case "ZodBoolean":
            return {
                type: "boolean",
            };
        case "ZodNullable":
            return Object.assign(Object.assign({}, serializeZodType(zt._def.innerType)), { nullable: true });
        case "ZodOptional":
            return Object.assign(Object.assign({}, serializeZodType(zt._def.innerType)), { optional: true });
        case "ZodAny":
            return {
                type: "any",
            };
        case "ZodRecord":
            return {
                type: "record",
                keyType: serializeZodType(zt._def.keyType),
                valueType: serializeZodType(zt._def.valueType),
            };
        case "ZodUnion":
            return {
                type: "union",
                options: zt._def.options.map((option) => serializeZodType(option)),
            };
        default:
            throw new Error(`Serialize 로직이 정의되지 않은 ZodType: ${zt._def.typeName}`);
    }
}
exports.serializeZodType = serializeZodType;
function zodTypeToTsTypeDef(zt) {
    if (zt._def.description) {
        return zt._def.description;
    }
    switch (zt._def.typeName) {
        case "ZodString":
            return "string";
        case "ZodNumber":
            return "number";
        case "ZodBoolean":
            return "boolean";
        case "ZodBigInt":
            return "bigint";
        case "ZodDate":
            return "date";
        case "ZodNull":
            return "null";
        case "ZodUndefined":
            return "undefined";
        case "ZodAny":
            return "any";
        case "ZodUnknown":
            return "unknown";
        case "ZodNever":
            return "never";
        case "ZodNullable":
            return zodTypeToTsTypeDef(zt._def.innerType) + " | null";
        case "ZodDefault":
            return zodTypeToTsTypeDef(zt._def.innerType);
        case "ZodRecord":
            return `{ [ key: ${zodTypeToTsTypeDef(zt._def.keyType)} ]: ${zodTypeToTsTypeDef(zt._def.valueType)}}`;
        case "ZodLiteral":
            if (typeof zt._def.value === "string") {
                return `"${zt._def.value}"`;
            }
            else {
                return `${zt._def.value}`;
            }
        case "ZodUnion":
            return `${zt._def.options
                .map((option) => zodTypeToTsTypeDef(option))
                .join(" | ")}`;
        case "ZodEnum":
            return `${zt._def.values.map((val) => `"${val}"`).join(" | ")}`;
        case "ZodArray":
            return `${zodTypeToTsTypeDef(zt._def.type)}[]`;
        case "ZodObject":
            const shape = zt.shape;
            return [
                "{",
                ...Object.keys(shape).map((key) => {
                    if (shape[key]._def.typeName === "ZodOptional") {
                        return `${key}?: ${zodTypeToTsTypeDef(shape[key]._def.innerType)},`;
                    }
                    else {
                        return `${key}: ${zodTypeToTsTypeDef(shape[key])},`;
                    }
                }),
                "}",
            ].join("\n");
        case "ZodOptional":
            return zodTypeToTsTypeDef(zt._def.innerType) + " | undefined";
        default:
            throw new Error(`처리되지 않은 ZodType ${zt._def.typeName}`);
    }
}
exports.zodTypeToTsTypeDef = zodTypeToTsTypeDef;
//# sourceMappingURL=code-converters.js.map