"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.Sonamu = void 0;
const chalk_1 = __importDefault(require("chalk"));
const zod_1 = require("zod");
const code_converters_1 = require("./code-converters");
const so_exceptions_1 = require("../exceptions/so-exceptions");
const smd_manager_1 = require("../smd/smd-manager");
const caster_1 = require("../api/caster");
const types_1 = require("../types/types");
const syncer_1 = require("../syncer/syncer");
const controller_1 = require("../utils/controller");
const db_1 = require("../database/db");
const base_model_1 = require("../database/base-model");
const utils_1 = require("../utils/utils");
const path_1 = __importDefault(require("path"));
const fs_1 = require("fs");
class SonamuClass {
    constructor() {
        this.isInitialized = false;
        this._apiRootPath = null;
        this._dbConfig = null;
        this._syncer = null;
        this._config = null;
    }
    set apiRootPath(apiRootPath) {
        this._apiRootPath = apiRootPath;
    }
    get apiRootPath() {
        if (this._apiRootPath === null) {
            throw new Error("Sonamu has not been initialized");
        }
        return this._apiRootPath;
    }
    get appRootPath() {
        return this.apiRootPath.split(path_1.default.sep).slice(0, -1).join(path_1.default.sep);
    }
    set dbConfig(dbConfig) {
        this._dbConfig = dbConfig;
    }
    get dbConfig() {
        if (this._dbConfig === null) {
            throw new Error("Sonamu has not been initialized");
        }
        return this._dbConfig;
    }
    set syncer(syncer) {
        this._syncer = syncer;
    }
    get syncer() {
        if (this._syncer === null) {
            throw new Error("Sonamu has not been initialized");
        }
        return this._syncer;
    }
    set config(config) {
        this._config = config;
    }
    get config() {
        if (this._config === null) {
            throw new Error("Sonamu has not been initialized");
        }
        return this._config;
    }
    init() {
        return __awaiter(this, void 0, void 0, function* () {
            if (this.isInitialized) {
                return;
            }
            this.apiRootPath = yield (0, utils_1.findApiRootPath)();
            const configPath = path_1.default.join(this.apiRootPath, "sonamu.config.json");
            if ((0, fs_1.existsSync)(configPath) === false) {
                throw new Error(`Cannot find sonamu.config.json in ${configPath}`);
            }
            this.config = JSON.parse((0, fs_1.readFileSync)(configPath).toString());
            // DB 로드
            this.dbConfig = yield db_1.DB.readKnexfile();
            console.log(chalk_1.default.green("DB Config Loaded!"));
            // SMD 로드
            yield smd_manager_1.SMDManager.autoload();
            // Syncer & Autoload: SMD / Models / Types / APIs
            console.time(chalk_1.default.cyan("autoload&sync:"));
            yield smd_manager_1.SMDManager.autoload();
            this.syncer = new syncer_1.Syncer();
            yield this.syncer.autoloadModels();
            yield this.syncer.autoloadTypes();
            yield this.syncer.autoloadApis();
            if ((0, controller_1.isLocal)()) {
                yield this.syncer.sync();
            }
            console.timeEnd(chalk_1.default.cyan("autoload&sync:"));
            this.isInitialized = true;
        });
    }
    withFastify(server, config) {
        return __awaiter(this, void 0, void 0, function* () {
            if (this.isInitialized === false) {
                yield this.init();
            }
            // 전체 라우팅 리스트
            server.get(`${this.config.route.prefix}/routes`, (_request, _reply) => __awaiter(this, void 0, void 0, function* () {
                return this.syncer.apis;
            }));
            // Healthcheck API
            server.get(`${this.config.route.prefix}/healthcheck`, (_request, _reply) => __awaiter(this, void 0, void 0, function* () {
                return "ok";
            }));
            // API 라우팅 등록
            this.syncer.apis.map((api) => {
                // model
                if (this.syncer.models[api.modelName] === undefined) {
                    throw new Error(`정의되지 않은 모델에 접근 ${api.modelName}`);
                }
                const model = this.syncer.models[api.modelName];
                // 파라미터 정보로 zod 스키마 빌드
                const ReqType = (0, code_converters_1.getZodObjectFromApi)(api, this.syncer.types);
                // route
                server.route({
                    method: api.options.httpMethod,
                    url: this.config.route.prefix + api.path,
                    handler: (request, reply) => __awaiter(this, void 0, void 0, function* () {
                        var _a, _b;
                        const which = api.options.httpMethod === "GET" ? "query" : "body";
                        let reqBody;
                        try {
                            reqBody = (0, caster_1.fastifyCaster)(ReqType).parse((_a = request[which]) !== null && _a !== void 0 ? _a : {});
                        }
                        catch (e) {
                            if (e instanceof zod_1.ZodError) {
                                // TODO: BadRequest 에러 핸들링 (ZodError issues를 humanize하여 출력하는 로직 필요)
                                throw new so_exceptions_1.BadRequestException(`${e.issues[0].message}`, e.errors);
                            }
                            else {
                                throw e;
                            }
                        }
                        const result = yield model[api.methodName].apply(model, api.parameters.map((param) => {
                            // Context 인젝션
                            if (types_1.ApiParamType.isContext(param.type)) {
                                return config.contextProvider({
                                    headers: request.headers,
                                    reply,
                                }, request, reply);
                            }
                            else {
                                return reqBody[param.name];
                            }
                        }));
                        reply.type((_b = api.options.contentType) !== null && _b !== void 0 ? _b : "application/json");
                        return result;
                    }),
                }); // END server.route
            });
        });
    }
    destroy() {
        return __awaiter(this, void 0, void 0, function* () {
            yield base_model_1.BaseModel.destroy();
        });
    }
}
exports.Sonamu = new SonamuClass();
//# sourceMappingURL=init.js.map