"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.BaseModel = exports.BaseModelClass = void 0;
const luxon_1 = require("luxon");
const lodash_1 = require("lodash");
const knex_on_duplicate_update_1 = require("./knex-plugins/knex-on-duplicate-update");
(0, knex_on_duplicate_update_1.attachOnDuplicateUpdate)();
const db_1 = require("./db");
const types_1 = require("../types/types");
const inflection_1 = require("inflection");
const chalk_1 = __importDefault(require("chalk"));
const upsert_builder_1 = require("./upsert-builder");
class BaseModelClass {
    constructor() {
        this.modelName = "Unknown";
    }
    /* DB 인스턴스 get, destroy */
    getDB(which) {
        return db_1.DB.getDB(which);
    }
    destroy() {
        return __awaiter(this, void 0, void 0, function* () {
            return db_1.DB.destroy();
        });
    }
    myNow(timestamp) {
        const dt = timestamp === undefined
            ? luxon_1.DateTime.local()
            : luxon_1.DateTime.fromSeconds(timestamp);
        return dt.toFormat("yyyy-MM-dd HH:mm:ss");
    }
    getInsertedIds(wdb, rows, tableName, unqKeyFields, chunkSize = 500) {
        return __awaiter(this, void 0, void 0, function* () {
            if (!wdb) {
                wdb = this.getDB("w");
            }
            let unqKeys;
            let whereInField, selectField;
            if (unqKeyFields.length > 1) {
                whereInField = wdb.raw(`CONCAT_WS('_', '${unqKeyFields.join(",")}')`);
                selectField = `${whereInField} as tmpUid`;
                unqKeys = rows.map((row) => unqKeyFields.map((field) => row[field]).join("_"));
            }
            else {
                whereInField = unqKeyFields[0];
                selectField = unqKeyFields[0];
                unqKeys = rows.map((row) => row[unqKeyFields[0]]);
            }
            const chunks = (0, lodash_1.chunk)(unqKeys, chunkSize);
            let resultIds = [];
            for (let chunk of chunks) {
                const dbRows = yield wdb(tableName)
                    .select("id", wdb.raw(selectField))
                    .whereIn(whereInField, chunk);
                resultIds = resultIds.concat(dbRows.map((dbRow) => parseInt(dbRow.id)));
            }
            return resultIds;
        });
    }
    useLoaders(db, rows, loaders) {
        return __awaiter(this, void 0, void 0, function* () {
            if (loaders.length === 0) {
                return rows;
            }
            for (let loader of loaders) {
                let subQ;
                let subRows;
                let toCol;
                const fromIds = rows.map((row) => row[loader.manyJoin.idField]);
                if (loader.manyJoin.through === undefined) {
                    // HasMany
                    const idColumn = `${loader.manyJoin.toTable}.${loader.manyJoin.toCol}`;
                    subQ = db(loader.manyJoin.toTable)
                        .whereIn(idColumn, fromIds)
                        .select([...loader.select, idColumn]);
                    // HasMany에서 OneJoin이 있는 경우
                    loader.oneJoins.map((join) => {
                        if (join.join == "inner") {
                            subQ.innerJoin(`${join.table} as ${join.as}`, this.getJoinClause(db, join));
                        }
                        else if (join.join == "outer") {
                            subQ.leftOuterJoin(`${join.table} as ${join.as}`, this.getJoinClause(db, join));
                        }
                    });
                    toCol = loader.manyJoin.toCol;
                }
                else {
                    // ManyToMany
                    const idColumn = `${loader.manyJoin.through.table}.${loader.manyJoin.through.fromCol}`;
                    subQ = db(loader.manyJoin.through.table)
                        .join(loader.manyJoin.toTable, `${loader.manyJoin.through.table}.${loader.manyJoin.through.toCol}`, `${loader.manyJoin.toTable}.${loader.manyJoin.toCol}`)
                        .whereIn(idColumn, fromIds)
                        .select((0, lodash_1.uniq)([...loader.select, idColumn]));
                    // ManyToMany에서 OneJoin이 있는 경우
                    loader.oneJoins.map((join) => {
                        if (join.join == "inner") {
                            subQ.innerJoin(`${join.table} as ${join.as}`, this.getJoinClause(db, join));
                        }
                        else if (join.join == "outer") {
                            subQ.leftOuterJoin(`${join.table} as ${join.as}`, this.getJoinClause(db, join));
                        }
                    });
                    toCol = loader.manyJoin.through.fromCol;
                }
                subRows = yield subQ;
                if (loader.loaders) {
                    // 추가 -Many 케이스가 있는 경우 recursion 처리
                    subRows = yield this.useLoaders(db, subRows, loader.loaders);
                }
                // 불러온 row들을 참조ID 기준으로 분류 배치
                const subRowGroups = (0, lodash_1.groupBy)(subRows, toCol);
                rows = rows.map((row) => {
                    var _a;
                    row[loader.as] = ((_a = subRowGroups[row[loader.manyJoin.idField]]) !== null && _a !== void 0 ? _a : []).map((r) => (0, lodash_1.omit)(r, toCol));
                    return row;
                });
            }
            return rows;
        });
    }
    hydrate(rows) {
        return rows.map((row) => {
            // nullable relation인 경우 관련된 필드가 전부 null로 생성되는 것 방지하는 코드
            const nestedKeys = Object.keys(row).filter((key) => key.includes("__"));
            const groups = (0, lodash_1.groupBy)(nestedKeys, (key) => key.split("__")[0]);
            const nullKeys = Object.keys(groups).filter((key) => groups[key].length > 1 &&
                groups[key].every((field) => row[field] === null));
            const hydrated = Object.keys(row).reduce((r, field) => {
                if (!field.includes("__")) {
                    if (Array.isArray(row[field]) && (0, lodash_1.isObject)(row[field][0])) {
                        r[field] = this.hydrate(row[field]);
                        return r;
                    }
                    else {
                        r[field] = row[field];
                        return r;
                    }
                }
                const parts = field.split("__");
                const objPath = parts[0] +
                    parts
                        .slice(1)
                        .map((part) => `[${part}]`)
                        .join("");
                (0, lodash_1.set)(r, objPath, row[field] && Array.isArray(row[field]) && (0, lodash_1.isObject)(row[field][0])
                    ? this.hydrate(row[field])
                    : row[field]);
                return r;
            }, {});
            nullKeys.map((nullKey) => (hydrated[nullKey] = null));
            return hydrated;
        });
    }
    runSubsetQuery({ params, baseTable, subset, subsetQuery, build, debug, db: _db, }) {
        var _a;
        return __awaiter(this, void 0, void 0, function* () {
            const db = _db !== null && _db !== void 0 ? _db : this.getDB(subset.startsWith("A") ? "w" : "r");
            baseTable = baseTable !== null && baseTable !== void 0 ? baseTable : (0, inflection_1.pluralize)((0, inflection_1.underscore)(this.modelName));
            const queryMode = (_a = params.queryMode) !== null && _a !== void 0 ? _a : (params.id !== undefined ? "list" : "both");
            const { select, virtual, joins, loaders } = subsetQuery;
            const qb = build({
                qb: db.from(baseTable),
                db,
                select,
                joins,
                virtual,
            });
            // join
            joins.map((join) => {
                if (join.join == "inner") {
                    qb.innerJoin(`${join.table} as ${join.as}`, this.getJoinClause(db, join));
                }
                else if (join.join == "outer") {
                    qb.leftOuterJoin(`${join.table} as ${join.as}`, this.getJoinClause(db, join));
                }
            });
            // listQuery
            const rows = yield (() => __awaiter(this, void 0, void 0, function* () {
                if (queryMode === "count") {
                    return [];
                }
                // limit, offset
                if (params.num !== 0) {
                    qb.limit(params.num);
                    qb.offset(params.num * (params.page - 1));
                }
                // select, rows
                const listQuery = qb.clone().select(select);
                let rows = yield listQuery;
                // debug: listQuery
                if (debug === true || debug === "list") {
                    console.debug("DEBUG: list query", chalk_1.default.blue(listQuery.toQuery().toString()));
                }
                rows = yield this.useLoaders(db, rows, loaders);
                rows = this.hydrate(rows);
                return rows;
            }))();
            // countQuery
            const total = yield (() => __awaiter(this, void 0, void 0, function* () {
                var _b, _c;
                if (queryMode === "list") {
                    return undefined;
                }
                const clonedQb = qb.clone().clear("order").clear("offset").clear("limit");
                const [, matched] = (_b = clonedQb
                    .toQuery()
                    .toLowerCase()
                    .match(/select (distinct .+) from/)) !== null && _b !== void 0 ? _b : [];
                const countQuery = matched
                    ? clonedQb
                        .clear("select")
                        .select(db.raw(`COUNT(${matched.split(",")[0]}) as total`))
                        .first()
                    : clonedQb.clear("select").count("*", { as: "total" }).first();
                const countRow = yield countQuery;
                // debug: countQuery
                if (debug === true || debug === "count") {
                    console.debug("DEBUG: count query", chalk_1.default.blue(countQuery.toQuery().toString()));
                }
                return (_c = countRow === null || countRow === void 0 ? void 0 : countRow.total) !== null && _c !== void 0 ? _c : 0;
            }))();
            return { rows, total, subsetQuery, qb };
        });
    }
    getJoinClause(db, join) {
        if (!(0, types_1.isCustomJoinClause)(join)) {
            return db.raw(`${join.from} = ${join.to}`);
        }
        else {
            return db.raw(join.custom);
        }
    }
    getUpsertBuilder() {
        return new upsert_builder_1.UpsertBuilder();
    }
}
exports.BaseModelClass = BaseModelClass;
exports.BaseModel = new BaseModelClass();
//# sourceMappingURL=base-model.js.map