import { Knex } from "knex";
import { GenMigrationCode, MigrationColumn, MigrationForeign, MigrationIndex, MigrationSet, MigrationSetAndJoinTable } from "../types/types";
import { Entity } from "./entity";
type MigratorMode = "dev" | "deploy";
export type MigratorOptions = {
    readonly mode: MigratorMode;
};
type MigrationCode = {
    name: string;
    path: string;
};
type ConnString = `${"mysql2"}://${string}@${string}:${number}/${string}`;
export type MigrationStatus = {
    codes: MigrationCode[];
    conns: {
        name: string;
        connKey: string;
        connString: ConnString;
        currentVersion: string;
        status: string | number;
        pending: string[];
    }[];
    preparedCodes: GenMigrationCode[];
};
export declare class Migrator {
    readonly mode: MigratorMode;
    targets: {
        compare?: Knex;
        pending: Knex;
        shadow: Knex;
        apply: Knex[];
    };
    constructor(options: MigratorOptions);
    getMigrationCodes(): Promise<{
        normal: MigrationCode[];
        onlyTs: MigrationCode[];
        onlyJs: MigrationCode[];
    }>;
    getStatus(): Promise<MigrationStatus>;
    runAction(action: "latest" | "rollback", targets: string[]): Promise<{
        connKey: string;
        batchNo: number;
        applied: string[];
    }[]>;
    delCodes(codeNames: string[]): Promise<number>;
    generatePreparedCodes(): Promise<number>;
    clearPendingList(): Promise<void>;
    check(): Promise<void>;
    run(): Promise<void>;
    rollback(): Promise<void>;
    cleanUpDist(force?: boolean): Promise<void>;
    runShadowTest(): Promise<{
        connKey: string;
        batchNo: number;
        applied: string[];
    }[]>;
    resetAll(): Promise<void>;
    compareMigrations(compareDB: Knex): Promise<GenMigrationCode[]>;
    getMigrationSetFromDB(compareDB: Knex, table: string): Promise<MigrationSet | null>;
    resolveDBColType(colType: string, colField: string): Pick<MigrationColumn, "type" | "unsigned" | "length" | "precision" | "scale">;
    readTable(compareDB: Knex, tableName: string): Promise<[DBColumn[], DBIndex[], DBForeign[]]>;
    getMigrationSetFromMD(entity: Entity): MigrationSetAndJoinTable;
    genColumnDefinitions(columns: MigrationColumn[]): string[];
    genIndexDefinitions(indexes: MigrationIndex[]): string[];
    genForeignDefinitions(table: string, foreigns: MigrationForeign[]): {
        up: string[];
        down: string[];
    };
    generateCreateCode_ColumnAndIndexes(table: string, columns: MigrationColumn[], indexes: MigrationIndex[]): GenMigrationCode;
    generateCreateCode_Foreign(table: string, foreigns: MigrationForeign[]): GenMigrationCode[];
    showMigrationSet(which: string, migrationSet: MigrationSet): void;
    generateAlterCode_ColumnAndIndexes(table: string, entityColumns: MigrationColumn[], entityIndexes: MigrationIndex[], dbColumns: MigrationColumn[], dbIndexes: MigrationIndex[]): GenMigrationCode[];
    getAlterColumnsTo(entityColumns: MigrationColumn[], dbColumns: MigrationColumn[]): {
        add: MigrationColumn[];
        drop: MigrationColumn[];
        alter: MigrationColumn[];
    };
    getAlterColumnLinesTo(columnsTo: ReturnType<Migrator["getAlterColumnsTo"]>, entityColumns: MigrationColumn[]): {
        add: {
            up: string[];
            down: string[];
        };
        drop: {
            up: string[];
            down: string[];
        };
        alter: {
            up: string[];
            down: string[];
        };
    };
    getAlterIndexesTo(entityIndexes: MigrationIndex[], dbIndexes: MigrationIndex[]): {
        add: MigrationIndex[];
        drop: MigrationIndex[];
    };
    getAlterIndexLinesTo(indexesTo: ReturnType<Migrator["getAlterIndexesTo"]>, columnsTo: ReturnType<Migrator["getAlterColumnsTo"]>): {
        add: {
            up: string[];
            down: string[];
        };
        drop: {
            up: string[];
            down: string[];
        };
    };
    generateAlterCode_Foreigns(table: string, entityForeigns: MigrationForeign[], dbForeigns: MigrationForeign[]): GenMigrationCode[];
    destroy(): Promise<void>;
}
type DBColumn = {
    Field: string;
    Type: string;
    Null: string;
    Key: string;
    Default: string | null;
    Extra: string;
};
type DBIndex = {
    Table: string;
    Non_unique: number;
    Key_name: string;
    Seq_in_index: number;
    Column_name: string;
    Collation: string | null;
    Cardinality: number | null;
    Sub_part: number | null;
    Packed: string | null;
    Null: string;
    Index_type: string;
    Comment: string;
    Index_comment: string;
    Visible: string;
    Expression: string | null;
};
type DBForeign = {
    keyName: string;
    from: string;
    referencesTable: string;
    referencesField: string;
    onDelete: string;
    onUpdate: string;
};
export {};
//# sourceMappingURL=migrator.d.ts.map