"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.Template__generated = void 0;
const lodash_1 = require("lodash");
const entity_manager_1 = require("../entity/entity-manager");
const code_converters_1 = require("../api/code-converters");
const base_template_1 = require("./base-template");
const utils_1 = require("../utils/utils");
class Template__generated extends base_template_1.Template {
    constructor() {
        super("generated");
    }
    getTargetAndPath() {
        return {
            target: "api/src/application",
            path: `sonamu.generated.ts`,
        };
    }
    render({}) {
        const entityIds = entity_manager_1.EntityManager.getAllIds();
        const entities = entityIds.map((id) => entity_manager_1.EntityManager.get(id));
        // 전체 SourceCode 생성
        const sourceCodes = entities
            .map((entity) => {
            return [
                this.getEnumsSourceCode(entity),
                this.getBaseSchemaSourceCode(entity),
                this.getBaseListParamsSourceCode(entity),
                this.getSubsetSourceCode(entity),
            ].filter(utils_1.nonNullable);
        })
            .flat();
        // Sort
        const LABEL_KEY_ORDER = [
            "Enums",
            "BaseSchema",
            "BaseListParams",
            "Subsets",
            "SubsetQueries",
        ];
        sourceCodes.sort((a, b) => {
            const [aKey] = a.label.split(":");
            const [bKey] = b.label.split(":");
            const aIndex = LABEL_KEY_ORDER.indexOf(aKey);
            const bIndex = LABEL_KEY_ORDER.indexOf(bKey);
            if (aIndex > bIndex) {
                return 1;
            }
            else if (aIndex < bIndex) {
                return -1;
            }
            else {
                return 0;
            }
        });
        const sourceCode = sourceCodes.reduce((result, ts) => {
            if (ts === null) {
                return result;
            }
            return {
                lines: [...result.lines, `// ${ts.label}`, ...ts.lines, ""],
                importKeys: (0, lodash_1.uniq)([...result.importKeys, ...ts.importKeys]),
            };
        }, {
            lines: [],
            importKeys: [],
        });
        // .types.ts의 타입을 참조하는 경우 순환참조(상호참조)가 발생하므로 타입을 가져와 인라인 처리
        const allTypeKeys = entities
            .map((entity) => Object.keys(entity.types))
            .flat();
        const cdImportKeys = sourceCode.importKeys.filter((importKey) => allTypeKeys.includes(importKey));
        if (cdImportKeys.length > 0) {
            const customScalarLines = cdImportKeys
                .map((importKey) => {
                const entity = entities.find((entity) => entity.types[importKey]);
                if (!entity) {
                    throw new Error(`ZodType not found ${importKey}`);
                }
                const zodType = entity.types[importKey];
                return [
                    `// CustomScalar: ${importKey}`,
                    `const ${importKey} = ${(0, code_converters_1.zodTypeToZodCode)(zodType)};`,
                    `type ${importKey} = z.infer<typeof ${importKey}>`,
                    "",
                ];
            })
                .flat();
            sourceCode.lines = [...customScalarLines, ...sourceCode.lines];
            sourceCode.importKeys = sourceCode.importKeys.filter((importKey) => !cdImportKeys.includes(importKey));
        }
        const body = sourceCode.lines.join("\n");
        // import
        const sonamuImports = [
            "zArrayable",
            "SQLDateTimeString",
            "SubsetQuery",
            "SonamuQueryMode",
        ].filter((mod) => body.includes(mod));
        return Object.assign(Object.assign({}, this.getTargetAndPath()), { body, importKeys: sourceCode.importKeys, customHeaders: [
                `import { z } from 'zod';`,
                `import { ${sonamuImports.join(",")} } from "sonamu";`,
            ] });
    }
    getEnumsSourceCode(entity) {
        if (Object.keys(entity.enumLabels).length === 0) {
            return null;
        }
        return {
            label: `Enums: ${entity.id}`,
            lines: [
                ...Object.entries(entity.enumLabels)
                    .map(([enumId, enumLabel]) => [
                    `export const ${enumId} = z.enum([${Object.keys(enumLabel).map((el) => `"${el}"`)}]).describe("${enumId}");`,
                    `export type ${enumId} = z.infer<typeof ${enumId}>`,
                    `export const ${enumId}Label = ${JSON.stringify(enumLabel)}`,
                ])
                    .flat(),
            ],
            importKeys: [],
        };
    }
    getBaseSchemaSourceCode(entity, importKeys = []) {
        const schemaName = `${entity.names.module}BaseSchema`;
        const propNode = {
            nodeType: "object",
            children: entity.props.map((prop) => {
                return {
                    nodeType: "plain",
                    prop,
                };
            }),
        };
        const schemaBody = (0, code_converters_1.propNodeToZodTypeDef)(propNode, importKeys);
        const lines = [
            `export const ${schemaName} = ${schemaBody}`,
            `export type ${schemaName} = z.infer<typeof ${schemaName}>`,
        ];
        return {
            label: `BaseSchema: ${entity.id}`,
            importKeys,
            lines,
        };
    }
    getBaseListParamsSourceCode(entity) {
        // Prop 없는 MD인 경우 생성 제외
        if (entity.props.length === 0) {
            return null;
        }
        else if (entity.parentId !== undefined) {
            return null;
        }
        const schemaName = `${entity.names.module}BaseListParams`;
        const filterProps = entity.props.filter((prop) => prop.toFilter === true);
        const propNodes = filterProps.map((prop) => {
            return {
                nodeType: "plain",
                prop,
                children: [],
            };
        });
        const importKeys = [];
        const filterBody = propNodes
            .map((propNode) => (0, code_converters_1.propNodeToZodTypeDef)(propNode, importKeys))
            .join("\n");
        const schemaBody = `
z.object({
  num: z.number().int().nonnegative(),
  page: z.number().int().min(1),
  search: ${entity.id}SearchField,
  keyword: z.string(),
  orderBy: ${entity.id}OrderBy,
  queryMode: SonamuQueryMode,
  id: zArrayable(z.number().int().positive()),${filterBody}
}).partial();
`.trim();
        const lines = [
            `export const ${schemaName} = ${schemaBody}`,
            `export type ${schemaName} = z.infer<typeof ${schemaName}>`,
        ];
        return {
            label: `BaseListParams: ${entity.id}`,
            importKeys,
            lines,
        };
    }
    getSubsetSourceCode(entity) {
        if (Object.keys(entity.subsets).length == 0) {
            return null;
        }
        else if (entity.parentId !== undefined) {
            return null;
        }
        const subsetKeys = Object.keys(entity.subsets);
        const importKeys = [];
        const lines = [
            ...subsetKeys
                .map((subsetKey) => {
                // 서브셋에서 FieldExpr[] 가져옴
                const fieldExprs = entity.subsets[subsetKey];
                // FieldExpr[]로 EntityPropNode[] 가져옴
                const propNodes = entity.fieldExprsToPropNodes(fieldExprs);
                const schemaName = `${entity.names.module}Subset${subsetKey}`;
                const propNode = {
                    nodeType: "object",
                    children: propNodes,
                };
                // EntityPropNode[]로 ZodTypeDef(string)을 가져옴
                const body = (0, code_converters_1.propNodeToZodTypeDef)(propNode, importKeys);
                return [
                    `export const ${schemaName} = ${body}`,
                    `export type ${schemaName} = z.infer<typeof ${schemaName}>`,
                ];
            })
                .flat(),
            `export type ${entity.names.module}SubsetMapping = {`,
            ...subsetKeys.map((subsetKey) => `  ${subsetKey}: ${entity.names.module}Subset${subsetKey};`),
            "}",
            `export const ${entity.names.module}SubsetKey = z.enum([${subsetKeys
                .map((k) => `"${k}"`)
                .join(",")}]);`,
            `export type ${entity.names.module}SubsetKey = z.infer<typeof ${entity.names.module}SubsetKey>`,
            "",
        ];
        return {
            label: `Subsets: ${entity.id}`,
            lines,
            importKeys: (0, lodash_1.uniq)(importKeys),
        };
    }
}
exports.Template__generated = Template__generated;
//# sourceMappingURL=generated.template.js.map