"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.Template__service = void 0;
const inflection_1 = require("inflection");
const lodash_1 = require("lodash");
const entity_manager_1 = require("../entity/entity-manager");
const types_1 = require("../types/types");
const code_converters_1 = require("../api/code-converters");
const base_template_1 = require("./base-template");
const sonamu_1 = require("../api/sonamu");
class Template__service extends base_template_1.Template {
    constructor() {
        super("service");
    }
    getTargetAndPath(names) {
        return {
            target: ":target/src/services",
            path: `${names.fs}/${names.fs}.service.ts`,
        };
    }
    render({ entityId }, apis) {
        const names = entity_manager_1.EntityManager.getNamesFromId(entityId);
        // 서비스 TypeSource
        const { lines, importKeys } = this.getTypeSource(apis);
        return Object.assign(Object.assign({}, this.getTargetAndPath(names)), { body: lines.join("\n"), importKeys: importKeys.filter((key) => ["ListResult"].includes(key) === false), customHeaders: [
                `import { z } from 'zod';`,
                `import qs from "qs";`,
                `import useSWR, { SWRResponse } from "swr";`,
                `import { fetch, ListResult, SWRError, SwrOptions, handleConditional, swrPostFetcher } from '../sonamu.shared';`,
            ] });
    }
    getTypeSource(apis) {
        const importKeys = [];
        // 제네릭에서 선언한 타입, importKeys에서 제외 필요
        let typeParamNames = [];
        const groups = (0, lodash_1.groupBy)(apis, (api) => api.modelName);
        const body = Object.keys(groups)
            .map((modelName) => {
            const methods = groups[modelName];
            const methodCodes = methods
                .map((api) => {
                // 컨텍스트 제외된 파라미터 리스트
                const paramsWithoutContext = api.parameters.filter((param) => !types_1.ApiParamType.isContext(param.type) &&
                    !types_1.ApiParamType.isRefKnex(param.type));
                // 파라미터 타입 정의
                const typeParamsDef = api.typeParameters
                    .map((typeParam) => {
                    return (0, code_converters_1.apiParamTypeToTsType)(typeParam, importKeys);
                })
                    .join(", ");
                typeParamNames = typeParamNames.concat(api.typeParameters.map((typeParam) => typeParam.id));
                // 파라미터 정의
                const paramsDef = (0, code_converters_1.apiParamToTsCode)(paramsWithoutContext, importKeys);
                // 리턴 타입 정의
                const returnTypeDef = (0, code_converters_1.apiParamTypeToTsType)((0, code_converters_1.unwrapPromiseOnce)(api.returnType), importKeys);
                // 페이로드 데이터 정의
                const payloadDef = `{ ${paramsWithoutContext
                    .map((param) => param.name)
                    .join(", ")} }`;
                return (0, lodash_1.sortBy)(api.options.clients, (client) => client === "swr" ? 0 : 1)
                    .map((client) => {
                    const apiBaseUrl = `${sonamu_1.Sonamu.config.route.prefix}${api.path}`;
                    switch (client) {
                        case "axios":
                            return this.renderAxios(api, apiBaseUrl, typeParamsDef, paramsDef, returnTypeDef, payloadDef);
                        case "axios-multipart":
                            return this.renderAxiosMultipart(api, apiBaseUrl, typeParamsDef, paramsDef, returnTypeDef, paramsWithoutContext);
                        case "swr":
                            return this.renderSwr(api, apiBaseUrl, typeParamsDef, paramsDef, returnTypeDef, payloadDef);
                        case "window-fetch":
                            return this.renderWindowFetch(api, apiBaseUrl, typeParamsDef, paramsDef, payloadDef);
                        case "socketio":
                        default:
                            return `// Not supported ${(0, inflection_1.camelize)(client, true)} yet.`;
                    }
                })
                    .join("\n");
            })
                .join("\n\n");
            return `export namespace ${modelName.replace(/Model$/, "Service")} {
${methodCodes}
}`;
        })
            .join("\n\n");
        return {
            lines: [body],
            importKeys: (0, lodash_1.difference)((0, lodash_1.uniq)(importKeys), typeParamNames),
        };
    }
    renderAxios(api, apiBaseUrl, typeParamsDef, paramsDef, returnTypeDef, payloadDef) {
        const methodNameAxios = api.options.resourceName
            ? "get" + (0, inflection_1.camelize)(api.options.resourceName)
            : api.methodName;
        if (api.options.httpMethod === "GET") {
            return `
export async function ${methodNameAxios}${typeParamsDef}(${paramsDef}): Promise<${returnTypeDef}> {
    return fetch({
      method: "GET",
      url: \`${apiBaseUrl}?\${qs.stringify(${payloadDef})}\`,
    });
}
    `.trim();
        }
        else {
            return `
export async function ${methodNameAxios}${typeParamsDef}(${paramsDef}): Promise<${returnTypeDef}> {
    return fetch({
      method: '${api.options.httpMethod}',
      url: \`${apiBaseUrl}\`,
      data: ${payloadDef},
    });
}
      `.trim();
        }
    }
    renderAxiosMultipart(api, apiBaseUrl, typeParamsDef, paramsDef, returnTypeDef, paramsWithoutContext) {
        const formDataDef = [
            'formData.append("file", file);',
            ...paramsWithoutContext.map((param) => `formData.append('${param.name}', String(${param.name}));`),
        ].join("\n");
        const paramsDefComma = paramsDef !== "" ? ", " : "";
        return `
export async function ${api.methodName}${typeParamsDef}(
  ${paramsDef}${paramsDefComma}
  file: File,
  onUploadProgress?: (pe:ProgressEvent) => void
  ): Promise<${returnTypeDef}> {
    const formData = new FormData();
    ${formDataDef}
    return fetch({
      method: 'POST',
      url: \`${apiBaseUrl}\`,
      headers: {
        "Content-Type": "multipart/form-data",
      },
      onUploadProgress,
      data: formData,
    });
  }
  `.trim();
    }
    renderSwr(api, apiBaseUrl, typeParamsDef, paramsDef, returnTypeDef, payloadDef) {
        const methodNameSwr = api.options.resourceName
            ? "use" + (0, inflection_1.camelize)(api.options.resourceName)
            : "use" + (0, inflection_1.camelize)(api.methodName);
        return `  export function ${(0, inflection_1.camelize)(methodNameSwr, true)}${typeParamsDef}(${[paramsDef, "options?: SwrOptions"]
            .filter((p) => p !== "")
            .join(",")}, ): SWRResponse<${returnTypeDef}, SWRError> {
    return useSWR(handleConditional([
      \`${apiBaseUrl}\`,
      ${payloadDef},
    ], options?.conditional)${api.options.httpMethod === "POST" ? ", swrPostFetcher" : ""});
  }`;
    }
    renderWindowFetch(api, apiBaseUrl, typeParamsDef, paramsDef, payloadDef) {
        return `
export async function ${api.methodName}${typeParamsDef}(${paramsDef}): Promise<Response> {
    return window.fetch(\`${apiBaseUrl}?\${qs.stringify(${payloadDef})}\`);
}
    `.trim();
    }
}
exports.Template__service = Template__service;
//# sourceMappingURL=service.template.js.map